import { useCallback, useMemo, useRef, useState } from "react";
import SwaggerClient from "swagger-client";
import { useAsync } from "../Async";

/**
 * @external Promise
 * @see {@link https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise Promise}
 */

/**
 * @typedef AsyncFunction
 * @returns Promise
 */

async function createAPIClient(options) {
  const client = await SwaggerClient(options);
  // update the server url to be our intended server
  client.spec.servers[0].url = options.server;
  return client;
}

export function useCreateAPIClient(options) {
  const create = useCallback(async () => {
    return await createAPIClient(options);
  }, [options]);
  return useAsync({ fcn: create });
}

const defaultUnpacker = (x) => x;

/**
 * @typedef UnpackerFunction
 * @param
 */

/**
 * @typedef UseApiMethodProperties
 * @type {object}
 * @property {object} value Object returned by the resolved promise of `fcn`
 * @property {function} wrapper Callable wrapper for the `fcn`.
 * @property {boolean} loading Is `true` once `fcn` has been called, and becomes `false` once `fcn` resolves to data or an error.
 * @property {boolean} dataReady Is `false` until `fcn` resolves to data.
 * @property {object} error Result of `fcn` resolving to an error.
 * @property {function} abort Reference to AbortController#abort method
 */

/**
 * Hook that wraps (by default Swagger) http clients with loading,
 * error, and cleanup functionality.
 *
 * @param {AsyncFunction} fcn Async function that will be invoked
 * @param {object} params URL search params and path params as defined by OpenApi spec
 * @param {object} options Additional options to provide to the client, including requestBody in e.g. POST requests
 * @param {boolean} abortOnReentry If true, aborts the http request when called while busy/loading.
 * @param {...object} rest Additional arguments to pass to useAsync internally
 * @returns {UseApiMethodProperties}
 */
export function useAPIMethod({
  fcn,
  params,
  options,
  abortOnReentry = true,
  unpacker = defaultUnpacker,
  ...rest
}) {
  const controller = useRef(new AbortController());

  const method = useCallback(
    async (callSiteParams, callSiteOptions) => {
      const response = await fcn(
        { ...params, ...callSiteParams },
        { signal: controller.current.signal, ...options, ...callSiteOptions }
      );
      const unpacked = unpacker(response.obj, response.status);
      return unpacked;
    },
    [fcn, options, params, unpacker]
  );

  const abort = useCallback(() => {
    controller.current.abort();
    controller.current = new AbortController();
    console.debug("...Aborted");
  }, []);
  const onReentrance = useCallback(() => {
    if (abortOnReentry) {
      abort();
    }
    // We want the wrapper to be invoked
    // again on reentry
    return false;
  }, [abortOnReentry, abort]);

  const asyncState = useAsync({
    fcn: method,
    onReentrance,
    onCleanup: abort,
    ...rest
  });

  return { ...asyncState, abort };
}

const isZeroOrPositiveInt = (val) => {
  return Number.isInteger(val) && val >= 0;
};

/**
 * Manages pagination state e.g. outside a table or url.
 *
 * Provides getting/setting for a single pagination object, but internally
 * manages changes for each attribute independently so that e.g. if there are
 * no changes then the object is unchanged.
 */
export const usePagination = ({
  initPage,
  initLimit,
  initTotalCount,
  rowsPerPageOptions
}) => {
  const [page, setPage] = useState(initPage ?? 0);
  const [_rowsPerPageOptions] = useState(rowsPerPageOptions ?? [20, 50]);
  const [limit, setLimit] = useState(initLimit ?? _rowsPerPageOptions[0]);
  const [totalCount, setTotalCount] = useState(initTotalCount ?? 0);

  /**
   * Returns current set of pagination values
   */
  const pagination = useMemo(() => {
    const val = {
      page,
      limit,
      rows: limit,
      totalCount,
      totalRecords: totalCount,
      rowsPerPageOptions: _rowsPerPageOptions
    };
    return val;
  }, [page, limit, totalCount, _rowsPerPageOptions]);

  /**
   * Sets pagination. Ignores negative or nonzero values.
   */
  const setPagination = useCallback(
    ({ page, limit, totalCount, rows, totalRecords }) => {
      if (isZeroOrPositiveInt(page)) {
        setPage(page);
      }
      if (isZeroOrPositiveInt(limit || rows)) {
        setLimit(limit || rows);
      }
      if (isZeroOrPositiveInt(totalCount || totalRecords)) {
        setTotalCount(totalCount || totalRecords);
      }
    },
    []
  );

  return {
    pagination,
    setPagination,
    setTotalCount,
    setLimit,
    setPage
  };
};
