import { renderHook, waitFor } from "@testing-library/react";
import { describe, it, vi, expect } from "vitest";
import { useAPIMethod, useCreateAPIClient } from "./API";
import { useAsync } from "../Async";

// These urls are intercepted by MSW
const options = {
  url: "https://petstore.com/spec",
  server: "https://petstore.com"
};

async function getClient() {
  const { result } = renderHook(() => useCreateAPIClient(options));
  await waitFor(() => expect(result.current.dataReady).toBe(true));
  return result.current.value;
}

describe("useCreateAPIClient", () => {
  it("builds API client from remote spec", async () => {
    const client = await getClient();
    expect(Object.keys(client.apis)).toEqual([
      "hosts",
      "echo",
      "basic",
      "auth"
    ]);
    expect(client.spec.servers[0].url).toBe(options.server);
  });

  it("is possible to use the api", async () => {
    const client = await getClient();
    const { result } = renderHook(() =>
      useAsync({ fcn: client.apis.hosts.list_hosts })
    );
    await waitFor(() => expect(result.current.dataReady).toBe(true));

    const { hostList, totalCount } = result.current.value.obj;
    expect(totalCount).toBe(997);
    expect(hostList[0].owner).toBe("bcanter0");
  });
});

describe("useAPIMethod", () => {
  it("correctly wraps the api method", async () => {
    const client = await getClient();
    const stub = vi.fn(client.apis.hosts.list_hosts);
    const { result } = renderHook(() => useAPIMethod({ fcn: stub }));
    await waitFor(() => expect(result.current.dataReady).toBe(true));
    expect(stub).toHaveBeenCalledTimes(1);

    const { hostList, totalCount } = result.current.value;
    expect(totalCount).toBe(997);
    expect(hostList[0].owner).toBe("bcanter0");
  });

  it("is possible to pass default parameters", async () => {
    const client = await getClient();
    const stub = vi.fn(client.apis.hosts.list_hosts);
    const params = { limit: 77 };
    const { result } = renderHook(() => useAPIMethod({ fcn: stub, params }));
    await waitFor(() => expect(result.current.dataReady).toBe(true));
    expect(stub).toHaveBeenCalledTimes(1);

    const { hostList, totalCount } = result.current.value;
    expect(totalCount).toBe(997);
    expect(hostList.length).toBe(77);
    expect(hostList[0].owner).toBe("bcanter0");
  });

  it("is possible to pass parameters and a request body at the call site", async () => {
    const client = await getClient();
    const stub = vi.fn(client.apis.echo.echo);
    const { result } = renderHook(() =>
      useAPIMethod({ fcn: stub, call: false })
    );
    await result.current.wrapper(
      { queryEcho: "hello" },
      { requestBody: { bodyEcho: "world" } }
    );
    await waitFor(() => expect(result.current.dataReady).toBe(true));
    expect(stub).toHaveBeenCalledTimes(1);

    const { queryEcho, bodyEcho } = result.current.value;
    expect(queryEcho).toBe("hello");
    expect(bodyEcho).toBe("world");
  });
});
