import { renderHook, waitFor } from "@testing-library/react";
import { act } from "react";
import { describe, test, vi, expect } from "vitest";
import { useAsync } from "./Async";
export function sleep(ms) {
  return new Promise((resolve) => setTimeout(resolve, ms));
}

async function simulate_network(x = "") {
  await sleep(100);
  return `bar${x}`;
}

async function simulate_network_failure() {
  await sleep(100);
  throw new Error("Something bad happened");
}

describe("useAsync", () => {
  test("initializes correctly", () => {
    const stub = vi.fn(simulate_network);
    const { result } = renderHook(() =>
      useAsync({ fcn: stub, init: "foo", call: false })
    );
    const { value, wrapper, loading, dataReady } = result.current;
    expect(value).toBe("foo");
    expect(typeof wrapper).toBe("function");
    expect(loading).toBe(false);
    expect(dataReady).toBe(false);
    expect(stub).not.toBeCalled();
  });

  test("calls function and updates state correctly", async () => {
    const stub = vi.fn(simulate_network);
    const { result } = renderHook(() => useAsync({ fcn: stub }));

    // calling function should trigger a loading state
    await waitFor(() =>
      expect(result.current).toMatchObject({
        value: null,
        loading: true,
        dataReady: false
      })
    );

    // once the date is loaded, the value should be updated
    // there is data ready for consumption
    await waitFor(() =>
      expect(result.current).toMatchObject({
        value: "bar",
        loading: false,
        dataReady: true
      })
    );

    // the function should have been called exactly one time
    expect(stub).toBeCalledTimes(1);

    // manually trigger the function again and supply some data
    await act(() => result.current.wrapper(2));

    // once the data is loaded, the value is updated
    await waitFor(() =>
      expect(result.current).toMatchObject({
        value: "bar2",
        loading: false,
        dataReady: true
      })
    );

    // at this point the function should have been called exactly two times
    expect(stub).toBeCalledTimes(2);
  });

  test("cleans up afterwards if called immediately", async () => {
    const fcnStub = vi.fn(simulate_network);
    const cleanupStub = vi.fn();

    const { unmount } = renderHook(() =>
      useAsync({
        fcn: fcnStub,
        init: "foo",
        call: true,
        onCleanup: cleanupStub
      })
    );

    expect(fcnStub).toBeCalledTimes(1);
    expect(cleanupStub).not.toBeCalled();

    unmount();

    expect(cleanupStub).toBeCalledTimes(1);
  });

  test("returns awaitable wrapper", async () => {
    const stub = vi.fn(simulate_network);
    const { result } = renderHook(() => useAsync({ fcn: stub, call: false }));

    // manually trigger the function
    await act(() => result.current.wrapper(3));

    // the value returned directly from the wrapper should match the hook value.
    await waitFor(() => {
      expect(result.current).toMatchObject({
        value: "bar3",
        loading: false,
        dataReady: true
      });
    });

    // at this point the function should have been called exactly one time
    expect(stub).toBeCalledTimes(1);
  });

  test("handles errors", async () => {
    const stub = vi.fn(simulate_network_failure);
    const { result } = renderHook(() => useAsync({ fcn: stub }));

    // calling function should trigger a loading state
    await waitFor(() =>
      expect(result.current).toMatchObject({
        value: null,
        loading: true,
        dataReady: false,
        error: null
      })
    );

    // raising an error should remove the loading state
    // but the data will not be updated
    await waitFor(() =>
      expect(result.current).toMatchObject({
        value: null,
        loading: false,
        dataReady: false
      })
    );

    // and the the resulting error object should be available.
    expect(result.current.error).toEqual(new Error("Something bad happened"));

    // calling the function again should initially reset the error state.
    result.current.wrapper();
    await waitFor(() =>
      expect(result.current).toMatchObject({
        value: null,
        loading: true,
        dataReady: false,
        error: null
      })
    );
  });
});
